"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FileServiceClient = exports.BrowserFileSystemRegistryImpl = void 0;
const tslib_1 = require("tslib");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const di_1 = require("@opensumi/di");
const ide_core_browser_1 = require("@opensumi/ide-core-browser");
const ide_core_common_1 = require("@opensumi/ide-core-common");
const electron_1 = require("@opensumi/ide-core-common/lib/electron");
const iterator_1 = require("@opensumi/monaco-editor-core/esm/vs/base/common/iterator");
const common_1 = require("../common");
const watcher_1 = require("./watcher");
let BrowserFileSystemRegistryImpl = class BrowserFileSystemRegistryImpl {
    constructor() {
        this.providers = new Map();
    }
    registerFileSystemProvider(provider) {
        const scheme = provider.scheme;
        this.providers.set(scheme, provider);
        return {
            dispose: () => {
                this.providers.delete(scheme);
            },
        };
    }
};
BrowserFileSystemRegistryImpl = tslib_1.__decorate([
    (0, di_1.Injectable)()
], BrowserFileSystemRegistryImpl);
exports.BrowserFileSystemRegistryImpl = BrowserFileSystemRegistryImpl;
let FileServiceClient = class FileServiceClient {
    constructor() {
        this.watcherWithSchemaMap = new Map();
        this.watcherDisposerMap = new Map();
        this.onFileChangedEmitter = new ide_core_browser_1.Emitter();
        this.onFileProviderChangedEmitter = new ide_core_browser_1.Emitter();
        this._onFilesChanged = new ide_core_browser_1.Emitter();
        this.onFilesChanged = this._onFilesChanged.event;
        this._onFileProviderChanged = new ide_core_browser_1.Emitter();
        this.onFileProviderChanged = this._onFileProviderChanged.event;
        this._onDidChangeFileSystemProviderRegistrations = new ide_core_browser_1.Emitter();
        this.onDidChangeFileSystemProviderRegistrations = this._onDidChangeFileSystemProviderRegistrations.event;
        this._onDidChangeFileSystemProviderCapabilities = new ide_core_browser_1.Emitter();
        this.onDidChangeFileSystemProviderCapabilities = this._onDidChangeFileSystemProviderCapabilities.event;
        this.filesExcludesMatcherList = [];
        this.watcherId = 0;
        this.toDisposable = new ide_core_browser_1.DisposableCollection();
        this.watchFileExcludes = [];
        this.watchFileExcludesMatcherList = [];
        this.filesExcludes = [];
        this.workspaceRoots = [];
        // 记录哪些 fsProviders 发生了变更
        this._providerChanged = new Set();
        this.fsProviders = new Map();
        this.userHomeDeferred = new ide_core_browser_1.Deferred();
        this.options = {
            encoding: 'utf8',
            overwrite: false,
            recursive: true,
            moveToTrash: true,
        };
        this.uriWatcherMap = new Map();
        this.toDisposable.push(this.onDidChangeFileSystemProviderRegistrations((e) => {
            // 只支持 file
            if (e.added && e.scheme === ide_core_common_1.Schemes.file) {
                this.doGetCurrentUserHome();
            }
        }));
    }
    async doGetCurrentUserHome() {
        const provider = await this.getProvider(ide_core_common_1.Schemes.file);
        const userHome = provider.getCurrentUserHome();
        this.userHomeDeferred.resolve(userHome);
    }
    handlesScheme(scheme) {
        return this.registry.providers.has(scheme) || this.fsProviders.has(scheme);
    }
    dispose() {
        return this.toDisposable.dispose();
    }
    /**
     * 直接先读文件，错误在后端抛出
     * @deprecated 方法在未来或许有变化
     * @param uri URI 地址
     */
    async resolveContent(uri, options) {
        const _uri = this.convertUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const rawContent = await provider.readFile(_uri.codeUri);
        const data = rawContent.data || rawContent;
        const buffer = ide_core_browser_1.BinaryBuffer.wrap(Uint8Array.from(data));
        return { content: buffer.toString(options === null || options === void 0 ? void 0 : options.encoding) };
    }
    async readFile(uri) {
        const _uri = this.convertUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const rawContent = await provider.readFile(_uri.codeUri);
        const data = rawContent.data || rawContent;
        const buffer = ide_core_browser_1.BinaryBuffer.wrap(Uint8Array.from(data));
        return { content: buffer };
    }
    async getFileStat(uri, withChildren = true) {
        const _uri = this.convertUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        try {
            const stat = await provider.stat(_uri.codeUri);
            if (!stat) {
                throw common_1.FileSystemError.FileNotFound(_uri.codeUri.toString(), 'File not found.');
            }
            return this.filterStat(stat, withChildren);
        }
        catch (err) {
            if (common_1.FileSystemError.FileNotFound.is(err)) {
                return undefined;
            }
        }
    }
    async setContent(file, content, options) {
        const _uri = this.convertUri(file.uri);
        const provider = await this.getProvider(_uri.scheme);
        const stat = await provider.stat(_uri.codeUri);
        if (!stat) {
            throw common_1.FileSystemError.FileNotFound(file.uri, 'File not found.');
        }
        if (stat.isDirectory) {
            throw common_1.FileSystemError.FileIsADirectory(file.uri, 'Cannot set the content.');
        }
        if (!(await this.isInSync(file, stat))) {
            throw this.createOutOfSyncError(file);
        }
        await provider.writeFile(_uri.codeUri, typeof content === 'string' ? ide_core_browser_1.BinaryBuffer.fromString(content).buffer : content, { create: false, overwrite: true, encoding: options === null || options === void 0 ? void 0 : options.encoding });
        const newStat = await provider.stat(_uri.codeUri);
        return newStat;
    }
    async updateContent(file, contentChanges, options) {
        const _uri = this.convertUri(file.uri);
        const provider = await this.getProvider(_uri.scheme);
        const stat = await provider.stat(_uri.codeUri);
        if (!stat) {
            throw common_1.FileSystemError.FileNotFound(file.uri, 'File not found.');
        }
        if (stat.isDirectory) {
            throw common_1.FileSystemError.FileIsADirectory(file.uri, 'Cannot set the content.');
        }
        if (!this.checkInSync(file, stat)) {
            throw this.createOutOfSyncError(file);
        }
        if (contentChanges.length === 0) {
            return stat;
        }
        const content = (await provider.readFile(_uri.codeUri));
        const newContent = this.applyContentChanges(ide_core_browser_1.BinaryBuffer.wrap(content).toString(options === null || options === void 0 ? void 0 : options.encoding), contentChanges);
        await provider.writeFile(_uri.codeUri, ide_core_browser_1.BinaryBuffer.fromString(newContent).buffer, {
            create: false,
            overwrite: true,
            encoding: options === null || options === void 0 ? void 0 : options.encoding,
        });
        const newStat = await provider.stat(_uri.codeUri);
        if (!newStat) {
            throw common_1.FileSystemError.FileNotFound(_uri.codeUri.toString(), 'File not found.');
        }
        return newStat;
    }
    async createFile(uri, options) {
        const _uri = this.convertUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const content = ide_core_browser_1.BinaryBuffer.fromString((options === null || options === void 0 ? void 0 : options.content) || '').buffer;
        let newStat = await provider.writeFile(_uri.codeUri, content, {
            create: true,
            overwrite: (options && options.overwrite) || false,
            encoding: options === null || options === void 0 ? void 0 : options.encoding,
        });
        newStat = newStat || (await provider.stat(_uri.codeUri));
        return newStat;
    }
    async createFolder(uri) {
        const _uri = this.convertUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const result = await provider.createDirectory(_uri.codeUri);
        if (result) {
            return result;
        }
        const stat = await provider.stat(_uri.codeUri);
        return stat;
    }
    async move(sourceUri, targetUri, options) {
        const _sourceUri = this.convertUri(sourceUri);
        const _targetUri = this.convertUri(targetUri);
        const provider = await this.getProvider(_sourceUri.scheme);
        const result = await provider.rename(_sourceUri.codeUri, _targetUri.codeUri, {
            overwrite: !!(options && options.overwrite),
        });
        if (result) {
            return result;
        }
        const stat = await provider.stat(_targetUri.codeUri);
        return stat;
    }
    async copy(sourceUri, targetUri, options) {
        const _sourceUri = this.convertUri(sourceUri);
        const _targetUri = this.convertUri(targetUri);
        const provider = await this.getProvider(_sourceUri.scheme);
        const overwrite = await this.doGetOverwrite(options);
        if (!(0, common_1.containsExtraFileMethod)(provider, 'copy')) {
            throw this.getErrorProvideNotSupport(_sourceUri.scheme, 'copy');
        }
        const result = await provider.copy(_sourceUri.codeUri, _targetUri.codeUri, {
            overwrite: !!overwrite,
        });
        if (result) {
            return result;
        }
        const stat = await provider.stat(_targetUri.codeUri);
        return stat;
    }
    async getFsPath(uri) {
        if (!uri.startsWith('file:/')) {
            return undefined;
        }
        else {
            return ide_core_browser_1.FileUri.fsPath(uri);
        }
    }
    fireFilesChange(event) {
        const changes = event.changes.map((change) => ({
            uri: change.uri,
            type: change.type,
        }));
        this._onFilesChanged.fire(changes);
        this.eventBus.fire(new ide_core_browser_1.FilesChangeEvent(changes));
    }
    // 添加监听文件
    async watchFileChanges(uri, excludes) {
        const _uri = this.convertUri(uri.toString());
        const originWatcher = this.uriWatcherMap.get(_uri.toString());
        if (originWatcher) {
            // 这里兼容重连逻辑，重连时 watcher 会 disposed，需要重新生成
            if (!originWatcher.isDisposed()) {
                return originWatcher;
            }
            else {
                this.uriWatcherMap.delete(_uri.toString());
            }
        }
        const id = this.watcherId++;
        const provider = await this.getProvider(_uri.scheme);
        const schemaWatchIdList = this.watcherWithSchemaMap.get(_uri.scheme) || [];
        const watcherId = await provider.watch(_uri.codeUri, {
            excludes,
        });
        this.watcherDisposerMap.set(id, {
            dispose: async () => {
                const watcher = this.uriWatcherMap.get(_uri.toString());
                await Promise.all([provider.unwatch && provider.unwatch(watcherId), watcher && watcher.dispose()]);
                this.uriWatcherMap.delete(_uri.toString());
            },
        });
        schemaWatchIdList.push(id);
        this.watcherWithSchemaMap.set(_uri.scheme, schemaWatchIdList);
        const watcher = new watcher_1.FileSystemWatcher({
            fileServiceClient: this,
            watchId: id,
            uri,
        });
        this.uriWatcherMap.set(_uri.toString(), watcher);
        return watcher;
    }
    async setWatchFileExcludes(excludes) {
        const provider = await this.getProvider(ide_core_common_1.Schemes.file);
        return await provider.setWatchFileExcludes(excludes);
    }
    async getWatchFileExcludes() {
        const provider = await this.getProvider(ide_core_common_1.Schemes.file);
        return await provider.getWatchFileExcludes();
    }
    async setFilesExcludes(excludes, roots) {
        this.filesExcludes = excludes;
        this.filesExcludesMatcherList = [];
        if (roots) {
            this.setWorkspaceRoots(roots);
        }
        this.updateExcludeMatcher();
    }
    async setWorkspaceRoots(roots) {
        this.workspaceRoots = roots;
        this.updateExcludeMatcher();
    }
    async unwatchFileChanges(watcherId) {
        const disposable = this.watcherDisposerMap.get(watcherId);
        if (!disposable || !disposable.dispose) {
            return;
        }
        this.watcherDisposerMap.delete(watcherId);
        return disposable.dispose();
    }
    async delete(uriString, options) {
        if (this.appConfig.isElectronRenderer && (options === null || options === void 0 ? void 0 : options.moveToTrash) !== false) {
            const uri = new ide_core_browser_1.URI(uriString);
            if (uri.scheme === ide_core_common_1.Schemes.file) {
                return this.injector.get(electron_1.IElectronMainUIService).moveToTrash(uri.codeUri.fsPath);
            }
        }
        const _uri = this.convertUri(uriString);
        const provider = await this.getProvider(_uri.scheme);
        await provider.stat(_uri.codeUri);
        await provider.delete(_uri.codeUri, {
            recursive: true,
            moveToTrash: await this.doGetMoveToTrash(options),
        });
    }
    async getEncoding(uri) {
        // FIXME: 临时修复方案 目前识别率太低，全部返回 UTF8
        return 'utf8';
    }
    // capabilities
    listCapabilities() {
        return iterator_1.Iterable.map(this.fsProviders, ([scheme, provider]) => ({ scheme, capabilities: provider.capabilities }));
    }
    // capabilities end
    registerProvider(scheme, provider) {
        if (this.fsProviders.has(scheme)) {
            throw new Error(`The file system provider for \`${scheme}\` already registered`);
        }
        const disposables = [];
        this.fsProviders.set(scheme, provider);
        this._onDidChangeFileSystemProviderRegistrations.fire({ added: true, scheme, provider });
        disposables.push({
            dispose: () => {
                this._onDidChangeFileSystemProviderRegistrations.fire({ added: false, scheme, provider });
                this.fsProviders.delete(scheme);
                this._providerChanged.add(scheme);
            },
        });
        if (provider.onDidChangeFile) {
            disposables.push(provider.onDidChangeFile((e) => this.fireFilesChange({ changes: e })));
        }
        this.toDisposable.push(provider.onDidChangeCapabilities(() => this._onDidChangeFileSystemProviderCapabilities.fire({ provider, scheme })));
        disposables.push({
            dispose: async () => {
                const promises = [];
                (this.watcherWithSchemaMap.get(scheme) || []).forEach((id) => promises.push(this.unwatchFileChanges(id)));
                await Promise.all(promises);
            },
        });
        this._providerChanged.add(scheme);
        this.onFileProviderChangedEmitter.fire(Array.from(this._providerChanged));
        this.toDisposable.pushAll(disposables);
        /**
         * 当外部注册了当前 Provider 的时候，暴露出去一个 dispose 供注册方去调用
         */
        const tempToDisable = new ide_core_browser_1.DisposableCollection();
        tempToDisable.pushAll(disposables);
        this._onFileProviderChanged.fire(Array.from(this._providerChanged));
        return tempToDisable;
    }
    async access(uri, mode = common_1.FileAccess.Constants.F_OK) {
        const _uri = this.convertUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        if (!(0, common_1.containsExtraFileMethod)(provider, 'access')) {
            throw this.getErrorProvideNotSupport(_uri.scheme, 'access');
        }
        return await provider.access(_uri.codeUri, mode);
    }
    // 这里需要 try catch 了
    async getFileType(uri) {
        const _uri = this.convertUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        if (!(0, common_1.containsExtraFileMethod)(provider, 'getFileType')) {
            throw this.getErrorProvideNotSupport(_uri.scheme, 'getFileType');
        }
        return await provider.getFileType(uri);
    }
    // FIXME: file scheme only?
    async getCurrentUserHome() {
        return this.userHomeDeferred.promise;
    }
    getErrorProvideNotSupport(scheme, funName) {
        return `Scheme ${scheme} not support this function: ${funName}.`;
    }
    /**
     * 提供一个方法让集成方对该方法进行复写。
     * 如双容器架构中对 IDE 容器读取不到的研发容器目录进行 scheme 替换，让插件提供提供的 fs-provider 去读取
     */
    convertUri(uri) {
        const _uri = new ide_core_browser_1.URI(uri);
        if (!_uri.scheme) {
            throw new Error(`Unsupported convert Uri with non-scheme Uri: ${uri}`);
        }
        return _uri;
    }
    updateExcludeMatcher() {
        this.filesExcludes.forEach((str) => {
            if (this.workspaceRoots.length > 0) {
                this.workspaceRoots.forEach((root) => {
                    const uri = new ide_core_browser_1.URI(root);
                    const pathStrWithExclude = uri.resolve(str).path.toString();
                    this.filesExcludesMatcherList.push((0, ide_core_browser_1.parseGlob)(pathStrWithExclude));
                });
            }
            else {
                this.filesExcludesMatcherList.push((0, ide_core_browser_1.parseGlob)(str));
            }
        });
    }
    async getProvider(scheme) {
        if (this._providerChanged.has(scheme)) {
            // 让相关插件启动完成 (3秒超时), 此处防止每次都发，仅在相关scheme被影响时才尝试激活插件
            await this.eventBus.fireAndAwait(new ide_core_browser_1.ExtensionActivateEvent({ topic: 'onFileSystem', data: scheme }), {
                timeout: 3000,
            });
            this._providerChanged.delete(scheme);
        }
        const provider = this.fsProviders.get(scheme);
        if (!provider) {
            throw new Error(`Not find ${scheme} provider.`);
        }
        return provider;
    }
    async isReadonly(uriString) {
        try {
            const uri = new ide_core_browser_1.URI(uriString);
            const provider = await this.getProvider(uri.scheme);
            const stat = (await provider.stat(this.convertUri(uriString).codeUri));
            return !!stat.readonly;
        }
        catch (e) {
            // 考虑到非 readonly 变readonly 的情况，相对于 readonly 变不 readonly 来说更为严重
            return false;
        }
    }
    isExclude(uriString) {
        const uri = new ide_core_browser_1.URI(uriString);
        return this.filesExcludesMatcherList.some((matcher) => matcher(uri.path.toString()));
    }
    filterStat(stat, withChildren = true) {
        if (!stat) {
            return;
        }
        if (this.isExclude(stat.uri)) {
            return;
        }
        // 这里传了 false 就走不到后面递归逻辑了
        if (stat.children && withChildren) {
            stat.children = this.filterStatChildren(stat.children);
        }
        return stat;
    }
    filterStatChildren(children) {
        const list = [];
        children.forEach((child) => {
            if (this.isExclude(child.uri)) {
                return false;
            }
            const state = this.filterStat(child);
            if (state) {
                list.push(state);
            }
        });
        return list;
    }
    applyContentChanges(content, contentChanges) {
        let document = vscode_languageserver_types_1.TextDocument.create('', '', 1, content);
        for (const change of contentChanges) {
            let newContent = change.text;
            if (change.range) {
                const start = document.offsetAt(change.range.start);
                const end = document.offsetAt(change.range.end);
                newContent = document.getText().substr(0, start) + change.text + document.getText().substr(end);
            }
            document = vscode_languageserver_types_1.TextDocument.create(document.uri, document.languageId, document.version, newContent);
        }
        return document.getText();
    }
    async isInSync(file, stat) {
        if (this.checkInSync(file, stat)) {
            return true;
        }
        return false;
    }
    checkInSync(file, stat) {
        return stat.lastModification === file.lastModification && stat.size === file.size;
    }
    createOutOfSyncError(file) {
        return common_1.FileSystemError.FileIsOutOfSync(file.uri);
    }
    async doGetEncoding(option) {
        return option && typeof option.encoding !== 'undefined' ? option.encoding : this.options.encoding;
    }
    async doGetOverwrite(option) {
        return option && typeof option.overwrite !== 'undefined' ? option.overwrite : this.options.overwrite;
    }
    async doGetRecursive(option) {
        return option && typeof option.recursive !== 'undefined' ? option.recursive : this.options.recursive;
    }
    async doGetMoveToTrash(option) {
        return option && typeof option.moveToTrash !== 'undefined' ? option.moveToTrash : this.options.moveToTrash;
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(common_1.IBrowserFileSystemRegistry),
    tslib_1.__metadata("design:type", BrowserFileSystemRegistryImpl)
], FileServiceClient.prototype, "registry", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(di_1.INJECTOR_TOKEN),
    tslib_1.__metadata("design:type", di_1.Injector)
], FileServiceClient.prototype, "injector", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_common_1.IEventBus),
    tslib_1.__metadata("design:type", Object)
], FileServiceClient.prototype, "eventBus", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.AppConfig),
    tslib_1.__metadata("design:type", Object)
], FileServiceClient.prototype, "appConfig", void 0);
FileServiceClient = tslib_1.__decorate([
    (0, di_1.Injectable)(),
    tslib_1.__metadata("design:paramtypes", [])
], FileServiceClient);
exports.FileServiceClient = FileServiceClient;
//# sourceMappingURL=file-service-client.js.map