"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DiskFileSystemProvider = void 0;
const tslib_1 = require("tslib");
const os_1 = tslib_1.__importDefault(require("os"));
const path_1 = tslib_1.__importDefault(require("path"));
const fse = tslib_1.__importStar(require("fs-extra"));
const trash_1 = tslib_1.__importDefault(require("trash"));
const write_file_atomic_1 = tslib_1.__importDefault(require("write-file-atomic"));
const di_1 = require("@opensumi/di");
const ide_connection_1 = require("@opensumi/ide-connection");
const ide_core_node_1 = require("@opensumi/ide-core-node");
const ide_core_node_2 = require("@opensumi/ide-core-node");
const common_1 = require("../common/");
const file_service_watcher_1 = require("./file-service-watcher");
const file_type_1 = require("./shared/file-type");
const { Path } = ide_core_node_1.path;
const UNSUPPORTED_NODE_MODULES_EXCLUDE = '**/node_modules/*/**';
const DEFAULT_NODE_MODULES_EXCLUDE = '**/node_modules/**';
let DiskFileSystemProvider = class DiskFileSystemProvider extends ide_connection_1.RPCService {
    constructor() {
        super();
        this.fileChangeEmitter = new ide_core_node_2.Emitter();
        this.onDidChangeFile = this.fileChangeEmitter.event;
        this.watcherCollection = new Map();
        this.watchFileExcludes = [];
        this._whenReadyDeferred = new ide_core_node_1.Deferred();
        this.isInitialized = false;
        this.ignoreNextChangesEvent = new Set();
        this.onDidChangeCapabilities = ide_core_node_2.Event.None;
        this.logger = this.loggerManager.getLogger(ide_core_node_1.SupportLogNamespace.Node);
        this.initWatchServer();
    }
    get whenReady() {
        return this._whenReadyDeferred.promise;
    }
    get capabilities() {
        if (!this._capabilities) {
            this._capabilities =
                2 /* FileSystemProviderCapabilities.FileReadWrite */ |
                    4 /* FileSystemProviderCapabilities.FileOpenReadWriteClose */ |
                    16 /* FileSystemProviderCapabilities.FileReadStream */ |
                    8 /* FileSystemProviderCapabilities.FileFolderCopy */ |
                    8192 /* FileSystemProviderCapabilities.FileWriteUnlock */;
            if (ide_core_node_2.isLinux) {
                this._capabilities |= 1024 /* FileSystemProviderCapabilities.PathCaseSensitive */;
            }
        }
        return this._capabilities;
    }
    dispose() {
        var _a;
        (_a = this.watcherServerDisposeCollection) === null || _a === void 0 ? void 0 : _a.dispose();
    }
    /**
     * @param {Uri} uri
     * @param {{ excludes: string[] }}
     * @memberof DiskFileSystemProvider
     */
    async watch(uri, options) {
        var _a;
        await this.whenReady;
        const _uri = ide_core_node_2.Uri.revive(uri);
        const id = await this.watcherServer.watchFileChanges(_uri.toString(), {
            excludes: (_a = options === null || options === void 0 ? void 0 : options.excludes) !== null && _a !== void 0 ? _a : [],
        });
        const disposable = {
            dispose: () => {
                this.watcherServer.unwatchFileChanges(id);
            },
        };
        this.watcherCollection.set(_uri.toString(), { id, options, disposable });
        return id;
    }
    unwatch(watcherId) {
        for (const [_uri, { id, disposable }] of this.watcherCollection) {
            if (watcherId === id) {
                disposable.dispose();
            }
        }
    }
    async stat(uri, options) {
        const _uri = ide_core_node_2.Uri.revive(uri);
        try {
            const stat = await this.doGetStat(_uri, 1, options);
            if (stat) {
                return stat;
            }
            throw common_1.FileSystemError.Unavailable(uri.path, 'Error occurred while getting the file stat.');
        }
        catch (e) {
            this.logger.error(e);
            throw e;
        }
    }
    async readDirectory(uri) {
        const _uri = ide_core_node_2.Uri.revive(uri);
        const result = [];
        try {
            const dirList = await fse.readdir(_uri.fsPath);
            dirList.forEach((name) => {
                const filePath = path_1.default.join(_uri.fsPath, name);
                result.push([name, this.getFileStatType(fse.statSync(filePath))]);
            });
            return result;
        }
        catch (e) {
            return result;
        }
    }
    async createDirectory(uri) {
        const _uri = ide_core_node_2.Uri.revive(uri);
        const stat = await this.doGetStat(_uri, 0);
        if (stat) {
            if (stat.isDirectory) {
                return stat;
            }
            throw common_1.FileSystemError.FileExists(uri.path, 'Error occurred while creating the directory: path is a file.');
        }
        await fse.ensureDir(ide_core_node_2.FileUri.fsPath(new ide_core_node_2.URI(_uri)));
        const newStat = await this.doGetStat(_uri, 1);
        if (newStat) {
            return newStat;
        }
        throw common_1.FileSystemError.Unavailable(uri.path, 'Error occurred while creating the directory.');
    }
    async readFile(uri, encoding = 'utf8') {
        const _uri = ide_core_node_2.Uri.revive(uri);
        try {
            const buffer = await fse.readFile(ide_core_node_2.FileUri.fsPath(new ide_core_node_2.URI(_uri)));
            return buffer;
        }
        catch (error) {
            if ((0, common_1.isErrnoException)(error)) {
                if (error.code === 'ENOENT') {
                    throw common_1.FileSystemError.FileNotFound(uri.path, 'Error occurred while reading file');
                }
                if (error.code === 'EISDIR') {
                    throw common_1.FileSystemError.FileIsADirectory(uri.path, 'Error occurred while reading file: path is a directory.');
                }
                if (error.code === 'EPERM') {
                    throw common_1.FileSystemError.FileIsNoPermissions(uri.path, 'Error occurred while reading file: path is a directory.');
                }
            }
            throw error;
        }
    }
    async writeFile(uri, content, options) {
        const _uri = ide_core_node_2.Uri.revive(uri);
        const exists = await this.access(uri);
        if (exists && !options.overwrite) {
            throw common_1.FileSystemError.FileExists(_uri.toString());
        }
        else if (!exists && !options.create) {
            throw common_1.FileSystemError.FileNotFound(_uri.toString());
        }
        // fileServiceNode调用不会转换，前传通信会转换
        const buffer = content instanceof Buffer ? content : Buffer.from(Uint8Array.from(content));
        if (options.create) {
            return await this.createFile(uri, { content: buffer });
        }
        try {
            this.ignoreNextChangesEvent.add(_uri.toString());
            await (0, write_file_atomic_1.default)(ide_core_node_2.FileUri.fsPath(new ide_core_node_2.URI(_uri)), buffer);
        }
        catch (e) {
            await fse.writeFile(ide_core_node_2.FileUri.fsPath(new ide_core_node_2.URI(_uri)), buffer);
            this.logger.warn('Error using writeFileAtomicSync, using fs instead.', e);
        }
        finally {
            this.ignoreNextChangesEvent.delete(_uri.toString());
        }
    }
    access(uri, mode = common_1.FileAccess.Constants.F_OK) {
        return fse
            .access(ide_core_node_2.FileUri.fsPath(ide_core_node_2.URI.from(uri)), mode)
            .then(() => true)
            .catch(() => false);
    }
    async delete(uri, options) {
        const _uri = ide_core_node_2.Uri.revive(uri);
        const stat = await this.doGetStat(_uri, 0);
        if (!stat) {
            throw common_1.FileSystemError.FileNotFound(uri.path);
        }
        if (!(0, ide_core_node_2.isUndefined)(options.recursive)) {
            this.logger.warn('DiskFileSystemProvider not support options.recursive!');
        }
        // Windows 10.
        // Deleting an empty directory throws `EPERM error` instead of `unlinkDir`.
        // https://github.com/paulmillr/chokidar/issues/566
        // Force moveToTrash
        const moveToTrash = !!options.moveToTrash;
        if (moveToTrash) {
            return (0, trash_1.default)([ide_core_node_2.FileUri.fsPath(new ide_core_node_2.URI(_uri))]);
        }
        else {
            const filePath = ide_core_node_2.FileUri.fsPath(new ide_core_node_2.URI(_uri));
            const outputRootPath = path_1.default.join(os_1.default.tmpdir(), (0, ide_core_node_2.uuid)());
            try {
                await new Promise((resolve, reject) => {
                    fse.rename(filePath, outputRootPath, async (error) => {
                        if (error) {
                            return reject(error);
                        }
                        resolve();
                    });
                });
                // There is no reason for the promise returned by this function not to resolve
                // as soon as the move is complete.  Clearing up the temporary files can be
                // done in the background.
                fse.remove(ide_core_node_2.FileUri.fsPath(outputRootPath));
            }
            catch (error) {
                return fse.remove(filePath);
            }
        }
    }
    async rename(sourceUri, targetUri, options) {
        const result = await this.doMove(sourceUri, targetUri, options);
        return result;
    }
    async copy(sourceUri, targetUri, options) {
        const _sourceUri = ide_core_node_2.Uri.revive(sourceUri);
        const _targetUri = ide_core_node_2.Uri.revive(targetUri);
        const [sourceStat, targetStat] = await Promise.all([this.doGetStat(_sourceUri, 0), this.doGetStat(_targetUri, 0)]);
        const { overwrite, recursive } = options;
        if (!sourceStat) {
            throw common_1.FileSystemError.FileNotFound(sourceUri.path);
        }
        if (targetStat && !overwrite) {
            throw common_1.FileSystemError.FileExists(targetUri.path, "Did you set the 'overwrite' flag to true?");
        }
        if (targetStat && targetStat.uri === sourceStat.uri) {
            throw common_1.FileSystemError.FileExists(targetUri.path, 'Cannot perform copy, source and destination are the same.');
        }
        await fse.copy(ide_core_node_2.FileUri.fsPath(_sourceUri.toString()), ide_core_node_2.FileUri.fsPath(_targetUri.toString()), {
            overwrite,
            recursive,
        });
        const newStat = await this.doGetStat(_targetUri, 1);
        if (newStat) {
            return newStat;
        }
        throw common_1.FileSystemError.FileNotFound(targetUri.path, `Error occurred while copying ${sourceUri} to ${targetUri}.`);
    }
    async getCurrentUserHome() {
        return this.stat(ide_core_node_2.FileUri.create(os_1.default.homedir()).codeUri);
    }
    // 出于通信成本的考虑，排除文件的逻辑必须放在node层（fs provider层，不同的fs实现的exclude应该不一样）
    setWatchFileExcludes(excludes) {
        let watchExcludes = excludes;
        if (excludes.includes(UNSUPPORTED_NODE_MODULES_EXCLUDE)) {
            const idx = watchExcludes.findIndex((v) => v === UNSUPPORTED_NODE_MODULES_EXCLUDE);
            watchExcludes = watchExcludes.splice(idx, 1, DEFAULT_NODE_MODULES_EXCLUDE);
        }
        // 每次调用之后都需要重新初始化 WatcherServer，保证最新的规则生效
        this.logger.log('Set watcher exclude:', watchExcludes);
        this.watchFileExcludes = watchExcludes;
        this.initWatchServer(this.watchFileExcludes);
    }
    getWatchFileExcludes() {
        return this.watchFileExcludes;
    }
    getWatchExcludes(excludes) {
        return Array.from(new Set(this.watchFileExcludes.concat(excludes || [])));
    }
    initWatchServer(excludes) {
        if (!this.injector) {
            return;
        }
        if (this.watcherServerDisposeCollection) {
            this.watcherServerDisposeCollection.dispose();
        }
        this.watcherServerDisposeCollection = new ide_core_node_2.DisposableCollection();
        this.watcherServer = this.injector.get(file_service_watcher_1.FileSystemWatcherServer, [excludes]);
        this.watcherServer.setClient({
            onDidFilesChanged: (events) => {
                if (events.changes.length > 0) {
                    const changes = events.changes.filter((c) => !this.ignoreNextChangesEvent.has(c.uri));
                    this.fileChangeEmitter.fire(changes);
                    if (Array.isArray(this.rpcClient)) {
                        this.rpcClient.forEach((client) => {
                            client.onDidFilesChanged({
                                changes,
                            });
                        });
                    }
                }
            },
        });
        this.watcherServerDisposeCollection.push({
            dispose: () => {
                this.watcherServer.dispose();
            },
        });
        if (this.isInitialized) {
            // 当服务已经初始化一次后，重新初始化时需要重新绑定原有的监听服务
            this.rewatch();
        }
        else {
            this._whenReadyDeferred.resolve();
        }
        this.isInitialized = true;
    }
    async rewatch() {
        let tasks = [];
        for (const [uri, { id, options }] of this.watcherCollection) {
            tasks.push({
                id,
                uri,
                options,
            });
        }
        // 需要针对缓存根据路径深度排序，防止过度监听
        tasks = tasks.sort((a, b) => Path.pathDepth(a.uri) - Path.pathDepth(b.uri));
        for (const { uri, options } of tasks) {
            await this.watch(ide_core_node_2.Uri.parse(uri), { excludes: this.getWatchExcludes(options === null || options === void 0 ? void 0 : options.excludes) });
        }
    }
    async createFile(uri, options) {
        const _uri = ide_core_node_2.Uri.revive(uri);
        const parentUri = new ide_core_node_2.URI(_uri).parent;
        const parentStat = await this.doGetStat(parentUri.codeUri, 0);
        if (!parentStat) {
            await fse.ensureDir(ide_core_node_2.FileUri.fsPath(parentUri));
        }
        await fse.writeFile(ide_core_node_2.FileUri.fsPath(_uri.toString()), options.content);
        const newStat = await this.doGetStat(_uri, 1);
        if (newStat) {
            return newStat;
        }
        throw common_1.FileSystemError.FileNotFound(uri.path, 'Error occurred while creating the file.');
    }
    /**
     * Return `true` if it's possible for this URI to have children.
     * It might not be possible to be certain because of permission problems or other filesystem errors.
     */
    async mayHaveChildren(uri) {
        /* If there's a problem reading the root directory. Assume it's not empty to avoid overwriting anything.  */
        try {
            const rootStat = await this.doGetStat(uri, 0);
            if (rootStat === undefined) {
                return true;
            }
            /* Not a directory.  */
            if (rootStat !== undefined && rootStat.isDirectory === false) {
                return false;
            }
        }
        catch (error) {
            return true;
        }
        /* If there's a problem with it's children then the directory must not be empty.  */
        try {
            const stat = await this.doGetStat(uri, 1);
            if (stat !== undefined && stat.children !== undefined) {
                return stat.children.length > 0;
            }
            else {
                return true;
            }
        }
        catch (error) {
            return true;
        }
    }
    async doMove(sourceUri, targetUri, options) {
        const _sourceUri = ide_core_node_2.Uri.revive(sourceUri);
        const _targetUri = ide_core_node_2.Uri.revive(targetUri);
        const [sourceStat, targetStat] = await Promise.all([this.doGetStat(_sourceUri, 1), this.doGetStat(_targetUri, 1)]);
        const isCapitalizedEqual = _sourceUri.toString().toLocaleUpperCase() === _targetUri.toString().toLocaleUpperCase();
        const { overwrite } = options;
        if (!sourceStat) {
            throw common_1.FileSystemError.FileNotFound(sourceUri.path);
        }
        if (targetStat && !overwrite) {
            throw common_1.FileSystemError.FileExists(targetUri.path, "Did you set the 'overwrite' flag to true?");
        }
        // Different types. Files <-> Directory.
        if (targetStat && sourceStat.isDirectory !== targetStat.isDirectory) {
            if (targetStat.isDirectory) {
                throw common_1.FileSystemError.FileIsADirectory(targetStat.uri, `Cannot move '${sourceStat.uri}' file to an existing location.`);
            }
            throw common_1.FileSystemError.FileNotADirectory(targetStat.uri, `Cannot move '${sourceStat.uri}' directory to an existing location.`);
        }
        const [sourceMightHaveChildren, targetMightHaveChildren] = await Promise.all([
            this.mayHaveChildren(_sourceUri),
            this.mayHaveChildren(_targetUri),
        ]);
        // Handling special Windows case when source and target resources are empty folders.
        // Source should be deleted and target should be touched.
        if (!isCapitalizedEqual &&
            overwrite &&
            targetStat &&
            targetStat.isDirectory &&
            sourceStat.isDirectory &&
            !sourceMightHaveChildren &&
            !targetMightHaveChildren) {
            // 当移动路径跟目标路径均存在文件，同时排除大写路径不等时才进入该逻辑
            // 核心解决在 Mac 等默认大小写不敏感系统中的文件移动问题
            // The value should be a Unix timestamp in seconds.
            // For example, `Date.now()` returns milliseconds, so it should be divided by `1000` before passing it in.
            const now = Date.now() / 1000;
            await fse.utimes(ide_core_node_2.FileUri.fsPath(_targetUri.toString()), now, now);
            await fse.rmdir(ide_core_node_2.FileUri.fsPath(_sourceUri.toString()));
            const newStat = await this.doGetStat(_targetUri, 1);
            if (newStat) {
                return newStat;
            }
            throw common_1.FileSystemError.FileNotFound(targetUri.path, `Error occurred when moving resource from '${sourceUri.toString()}' to '${targetUri.toString()}'.`);
        }
        else if (overwrite &&
            targetStat &&
            targetStat.isDirectory &&
            sourceStat.isDirectory &&
            !targetMightHaveChildren &&
            sourceMightHaveChildren) {
            // Copy source to target, since target is empty. Then wipe the source content.
            const newStat = await this.copy(sourceUri, targetUri, { overwrite });
            await this.delete(sourceUri, { moveToTrash: false });
            return newStat;
        }
        else {
            await fse.move(ide_core_node_2.FileUri.fsPath(_sourceUri.toString()), ide_core_node_2.FileUri.fsPath(_targetUri.toString()), { overwrite });
            const stat = await this.doGetStat(_targetUri, 1, {
                throwError: true,
            });
            if (stat) {
                return stat;
            }
            else {
                // never reached
                throw common_1.FileSystemError.FileNotFound(targetUri.path, `Error occurred when moving resource from '${sourceUri.toString()}' to '${targetUri.toString()}'.`);
            }
        }
    }
    async doGetStat(uri, depth, options) {
        try {
            const filePath = uri.fsPath;
            const lstat = await fse.lstat(filePath);
            if (lstat.isSymbolicLink()) {
                let realPath;
                try {
                    realPath = await fse.realpath(ide_core_node_2.FileUri.fsPath(new ide_core_node_2.URI(uri)));
                }
                catch (e) {
                    return undefined;
                }
                const stat = await fse.stat(filePath);
                const realURI = ide_core_node_2.FileUri.create(realPath);
                const realStat = await fse.lstat(realPath);
                let realStatData;
                if (stat.isDirectory()) {
                    realStatData = await this.doCreateDirectoryStat(realURI.codeUri, realStat, depth);
                }
                else {
                    realStatData = await this.doCreateFileStat(realURI.codeUri, realStat);
                }
                return Object.assign(Object.assign({}, realStatData), { realUri: realStatData.uri, type: common_1.FileType.SymbolicLink, isSymbolicLink: true, uri: uri.toString() });
            }
            else {
                if (lstat.isDirectory()) {
                    return await this.doCreateDirectoryStat(uri, lstat, depth);
                }
                const fileStat = await this.doCreateFileStat(uri, lstat);
                return fileStat;
            }
        }
        catch (error) {
            if (options === null || options === void 0 ? void 0 : options.throwError) {
                (0, common_1.handleError)(error);
            }
            if ((0, common_1.isErrnoException)(error)) {
                if (error.code === 'ENOENT' || error.code === 'EACCES' || error.code === 'EBUSY' || error.code === 'EPERM') {
                    return undefined;
                }
            }
            throw error;
        }
    }
    async doCreateFileStat(uri, stat) {
        return {
            uri: uri.toString(),
            lastModification: stat.mtime.getTime(),
            createTime: stat.ctime.getTime(),
            isSymbolicLink: stat.isSymbolicLink(),
            isDirectory: stat.isDirectory(),
            size: stat.size,
            type: this.getFileStatType(stat),
        };
    }
    getFileStatType(stat) {
        if (stat.isDirectory()) {
            return common_1.FileType.Directory;
        }
        if (stat.isFile()) {
            return common_1.FileType.File;
        }
        if (stat.isSymbolicLink()) {
            return common_1.FileType.SymbolicLink;
        }
        return common_1.FileType.Unknown;
    }
    async doCreateDirectoryStat(uri, stat, depth) {
        const children = depth > 0 ? await this.doGetChildren(uri, depth) : [];
        return {
            uri: uri.toString(),
            lastModification: stat.mtime.getTime(),
            createTime: stat.ctime.getTime(),
            isDirectory: true,
            isSymbolicLink: stat.isSymbolicLink(),
            children,
            type: common_1.FileType.Directory,
        };
    }
    async doGetChildren(uri, depth) {
        const _uri = new ide_core_node_2.URI(uri);
        const files = await fse.readdir(ide_core_node_2.FileUri.fsPath(_uri));
        const children = await Promise.all(files.map((fileName) => _uri.resolve(fileName)).map((childUri) => this.doGetStat(childUri.codeUri, depth - 1)));
        return children.filter(common_1.notEmpty);
    }
    async getFileType(uri) {
        return await (0, file_type_1.getFileType)(uri);
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(di_1.INJECTOR_TOKEN),
    tslib_1.__metadata("design:type", di_1.Injector)
], DiskFileSystemProvider.prototype, "injector", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_node_1.ILogServiceManager),
    tslib_1.__metadata("design:type", Object)
], DiskFileSystemProvider.prototype, "loggerManager", void 0);
DiskFileSystemProvider = tslib_1.__decorate([
    (0, di_1.Injectable)({ multiple: true }),
    tslib_1.__metadata("design:paramtypes", [])
], DiskFileSystemProvider);
exports.DiskFileSystemProvider = DiskFileSystemProvider;
//# sourceMappingURL=disk-file-system.provider.js.map