"use strict";
var FileSystemWatcherServer_1;
Object.defineProperty(exports, "__esModule", { value: true });
exports.FileSystemWatcherServer = void 0;
const tslib_1 = require("tslib");
const path_1 = tslib_1.__importDefault(require("path"));
const watcher_1 = tslib_1.__importDefault(require("@parcel/watcher"));
const fs_extra_1 = tslib_1.__importDefault(require("fs-extra"));
const debounce_1 = tslib_1.__importDefault(require("lodash/debounce"));
const uniqBy_1 = tslib_1.__importDefault(require("lodash/uniqBy"));
const di_1 = require("@opensumi/di");
const ide_core_node_1 = require("@opensumi/ide-core-node");
const common_1 = require("../common");
const file_change_collection_1 = require("./file-change-collection");
let FileSystemWatcherServer = FileSystemWatcherServer_1 = class FileSystemWatcherServer {
    constructor(excludes = []) {
        this.excludes = excludes;
        this.WATCHER_HANDLERS = new Map();
        this.watcherOptions = new Map();
        this.toDispose = new ide_core_node_1.DisposableCollection(ide_core_node_1.Disposable.create(() => this.setClient(undefined)));
        this.changes = new file_change_collection_1.FileChangeCollection();
        /**
         * Fires file changes to clients.
         * It is debounced in the case if the filesystem is spamming to avoid overwhelming clients with events.
         */
        this.fireDidFilesChanged = (0, debounce_1.default)(() => this.doFireDidFilesChanged(), 100);
        this.logger = this.loggerManager.getLogger(ide_core_node_1.SupportLogNamespace.Node);
    }
    dispose() {
        this.toDispose.dispose();
        this.WATCHER_HANDLERS.clear();
    }
    /**
     * 查找某个路径是否已被监听
     * @param watcherPath
     */
    checkIsAlreadyWatched(watcherPath) {
        for (const [watcherId, watcher] of this.WATCHER_HANDLERS) {
            if (watcherPath.indexOf(watcher.path) === 0) {
                return watcherId;
            }
        }
    }
    /**
     * 如果监听路径不存在，则会监听父目录
     * @param uri 要监听的路径
     * @param options
     * @returns
     */
    async watchFileChanges(uri, options) {
        const basePath = ide_core_node_1.FileUri.fsPath(uri);
        const exist = await fs_extra_1.default.pathExists(basePath);
        let watcherId = this.checkIsAlreadyWatched(basePath);
        if (watcherId) {
            return watcherId;
        }
        watcherId = FileSystemWatcherServer_1.WATCHER_SEQUENCE++;
        const toDisposeWatcher = new ide_core_node_1.DisposableCollection();
        let watchPath;
        if (exist) {
            const stat = await fs_extra_1.default.lstatSync(basePath);
            if (stat && stat.isDirectory()) {
                watchPath = basePath;
            }
            else {
                watchPath = await this.lookup(basePath);
            }
        }
        else {
            watchPath = await this.lookup(basePath);
        }
        this.logger.log('Starting watching:', watchPath, options);
        const handler = (err, events) => {
            if (err) {
                this.logger.error(`Watching ${watchPath} error: `, err);
                return;
            }
            events = this.trimChangeEvent(events);
            for (const event of events) {
                if (event.type === 'create') {
                    this.pushAdded(event.path);
                }
                if (event.type === 'delete') {
                    this.pushDeleted(event.path);
                }
                if (event.type === 'update') {
                    this.pushUpdated(event.path);
                }
            }
        };
        this.WATCHER_HANDLERS.set(watcherId, {
            path: watchPath,
            disposable: toDisposeWatcher,
            handlers: [handler],
        });
        toDisposeWatcher.push(ide_core_node_1.Disposable.create(() => this.WATCHER_HANDLERS.delete(watcherId)));
        toDisposeWatcher.push(await this.start(watcherId, watchPath, options));
        this.toDispose.push(toDisposeWatcher);
        return watcherId;
    }
    /**
     * 向上查找存在的目录
     * 默认向上查找 5 层，避免造成较大的目录监听带来的性能问题
     * 当前框架内所有配置文件可能存在的路径层级均不超过 5 层
     * @param path 监听的文件路径
     * @param count 向上查找层级
     */
    async lookup(path, count = 5) {
        let uri = path_1.default.dirname(path);
        let times = 0;
        while (!(await fs_extra_1.default.pathExists(uri)) && times < count) {
            uri = path_1.default.dirname(uri);
            times++;
        }
        if (await fs_extra_1.default.pathExists(uri)) {
            return uri;
        }
        else {
            return '';
        }
    }
    /**
     * 过滤 `write-file-atomic` 写入生成的临时文件
     * @param events
     */
    trimChangeEvent(events) {
        events = events.filter((event) => {
            if (event.path) {
                if (this.isTempFile(event.path)) {
                    // write-file-atomic 源文件xxx.xx 对应的临时文件为 xxx.xx.22243434
                    // 这类文件的更新应当完全隐藏掉
                    return false;
                }
            }
            return true;
        });
        return events;
    }
    isTempFile(path) {
        return /\.\d{7}\d+$/.test(path);
    }
    getDefaultWatchExclude() {
        return ['**/.git/objects/**', '**/.git/subtree-cache/**', '**/node_modules/**/*', '**/.hg/store/**'];
    }
    async start(watcherId, basePath, rawOptions) {
        const disposables = new ide_core_node_1.DisposableCollection();
        if (!(await fs_extra_1.default.pathExists(basePath))) {
            return disposables;
        }
        const realPath = await fs_extra_1.default.realpath(basePath);
        const tryWatchDir = async (maxRetries = 3, retryDelay = 1000) => {
            for (let times = 0; times < maxRetries; times++) {
                try {
                    return await watcher_1.default.subscribe(realPath, (err, events) => {
                        var _a;
                        // 对于超过 5000 数量的 events 做屏蔽优化，避免潜在的卡死问题
                        if (events.length > 5000) {
                            // FIXME: 研究此处屏蔽的影响，考虑下阈值应该设置多少，或者更加优雅的方式
                            return;
                        }
                        const handlers = (_a = this.WATCHER_HANDLERS.get(watcherId)) === null || _a === void 0 ? void 0 : _a.handlers;
                        if (!handlers) {
                            return;
                        }
                        for (const handler of handlers) {
                            handler(err, events);
                        }
                    }, {
                        backend: FileSystemWatcherServer_1.PARCEL_WATCHER_BACKEND,
                        ignore: this.excludes.concat((rawOptions === null || rawOptions === void 0 ? void 0 : rawOptions.excludes) || this.getDefaultWatchExclude()),
                    });
                }
                catch (e) {
                    // Watcher 启动失败，尝试重试
                    this.logger.error('watcher subscribe failed ', e, ' try times ', times);
                    await new Promise((resolve) => {
                        setTimeout(resolve, retryDelay);
                    });
                }
            }
            // 经过若干次的尝试后，Parcel Watcher 依然启动失败，此时就不再尝试重试
            this.logger.error(`watcher subscribe finally failed after ${maxRetries} times`);
            return undefined; // watch 失败则返回 undefined
        };
        /**
         * 由于 parcel/watcher 在 Linux 下存在内存越界访问问题触发了 sigsegv 导致 crash，所以在 Linux 下仍旧使用 nsfw
         * 社区相关 issue: https://github.com/parcel-bundler/watcher/issues/49
         * 后续这里的 watcher 模块需要重构掉，先暂时这样处理
         *
         * 代码来自 issue: https://github.com/opensumi/core/pull/1437/files?diff=split&w=0#diff-9de963117a88a70d7c58974bf2b092c61a196d6eef719846d78ca5c9d100b796 的旧代码处理
         */
        if (this.isEnableNSFW()) {
            const nsfw = await this.withNSFWModule();
            const watcher = await nsfw(realPath, (events) => this.handleNSFWEvents(events, watcherId), {
                errorCallback: (error) => {
                    // see https://github.com/atom/github/issues/342
                    this.unwatchFileChanges(watcherId);
                },
            });
            await watcher.start();
            disposables.push(ide_core_node_1.Disposable.create(async () => {
                this.watcherOptions.delete(watcherId);
                await watcher.stop();
            }));
            const excludes = this.excludes.concat((rawOptions === null || rawOptions === void 0 ? void 0 : rawOptions.excludes) || this.getDefaultWatchExclude());
            this.watcherOptions.set(watcherId, {
                excludesPattern: excludes.map((pattern) => (0, ide_core_node_1.parseGlob)(pattern)),
                excludes,
            });
        }
        else {
            const hanlder = await tryWatchDir();
            if (hanlder) {
                // watch 成功才加入 disposables，否则也就无需 dispose
                disposables.push(ide_core_node_1.Disposable.create(async () => {
                    if (hanlder) {
                        await hanlder.unsubscribe();
                    }
                }));
            }
        }
        return disposables;
    }
    unwatchFileChanges(watcherId) {
        const watcher = this.WATCHER_HANDLERS.get(watcherId);
        if (watcher) {
            this.WATCHER_HANDLERS.delete(watcherId);
            watcher.disposable.dispose();
        }
        return Promise.resolve();
    }
    setClient(client) {
        if (client && this.toDispose.disposed) {
            return;
        }
        this.client = client;
    }
    /**
     * @deprecated
     * 主要是用来跳过 jest 测试
     */
    isEnableNSFW() {
        return ide_core_node_1.isLinux;
    }
    async handleNSFWEvents(events, watcherId) {
        if (events.length > 5000) {
            return;
        }
        const isIgnored = (watcherId, path) => {
            const options = this.watcherOptions.get(watcherId);
            if (!options || !options.excludes || options.excludes.length < 1) {
                return false;
            }
            return options.excludesPattern.some((match) => match(path));
        };
        const filterEvents = events.filter((event) => {
            // 如果是 RENAME，不会产生临时文件
            if (event.action === common_1.INsfw.actions.RENAMED) {
                return true;
            }
            return !this.isTempFile(event.file);
        });
        // 合并下事件，由于 resolvePath 耗时较久，这里只用当前事件路径及文件名去重，后续处理事件再获取真实路径
        const mergedEvents = (0, uniqBy_1.default)(filterEvents, (event) => {
            if (event.action === common_1.INsfw.actions.RENAMED) {
                const deletedPath = path_1.default.join(event.directory, event.oldFile);
                const newPath = path_1.default.join(event.newDirectory || event.directory, event.newFile);
                return deletedPath + newPath;
            }
            return event.action + path_1.default.join(event.directory, event.file);
        });
        for (const event of mergedEvents) {
            if (event.action === common_1.INsfw.actions.RENAMED) {
                const deletedPath = this.resolvePath(event.directory, event.oldFile);
                if (isIgnored(watcherId, deletedPath)) {
                    continue;
                }
                this.pushDeleted(deletedPath);
                if (event.newDirectory) {
                    const path = this.resolvePath(event.newDirectory, event.newFile);
                    if (isIgnored(watcherId, path)) {
                        continue;
                    }
                    this.pushAdded(path);
                }
                else {
                    const path = this.resolvePath(event.directory, event.newFile);
                    if (isIgnored(watcherId, path)) {
                        continue;
                    }
                    this.pushAdded(path);
                }
            }
            else {
                const path = this.resolvePath(event.directory, event.file);
                if (isIgnored(watcherId, path)) {
                    continue;
                }
                if (event.action === common_1.INsfw.actions.CREATED) {
                    this.pushAdded(path);
                }
                if (event.action === common_1.INsfw.actions.DELETED) {
                    this.pushDeleted(path);
                }
                if (event.action === common_1.INsfw.actions.MODIFIED) {
                    this.pushUpdated(path);
                }
            }
        }
    }
    async withNSFWModule() {
        return require('nsfw');
    }
    pushAdded(path) {
        this.pushFileChange(path, common_1.FileChangeType.ADDED);
    }
    pushUpdated(path) {
        this.pushFileChange(path, common_1.FileChangeType.UPDATED);
    }
    pushDeleted(path) {
        this.pushFileChange(path, common_1.FileChangeType.DELETED);
    }
    pushFileChange(path, type) {
        const uri = ide_core_node_1.FileUri.create(path).toString();
        this.changes.push({ uri, type });
        this.fireDidFilesChanged();
    }
    resolvePath(directory, file) {
        const path = path_1.default.join(directory, file);
        // 如果是 linux 则获取一下真实 path，以防返回的是软连路径被过滤
        if (ide_core_node_1.isLinux) {
            try {
                return fs_extra_1.default.realpathSync.native(path);
            }
            catch (_e) {
                try {
                    // file does not exist try to resolve directory
                    return path_1.default.join(fs_extra_1.default.realpathSync.native(directory), file);
                }
                catch (_e) {
                    // directory does not exist fall back to symlink
                    return path;
                }
            }
        }
        return path;
    }
    doFireDidFilesChanged() {
        const changes = this.changes.values();
        this.changes = new file_change_collection_1.FileChangeCollection();
        const event = { changes };
        if (this.client) {
            this.client.onDidFilesChanged(event);
        }
    }
};
FileSystemWatcherServer.PARCEL_WATCHER_BACKEND = ide_core_node_1.isWindows ? 'windows' : ide_core_node_1.isLinux ? 'inotify' : 'fs-events';
FileSystemWatcherServer.WATCHER_SEQUENCE = 1;
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_node_1.ILogServiceManager),
    tslib_1.__metadata("design:type", Object)
], FileSystemWatcherServer.prototype, "loggerManager", void 0);
FileSystemWatcherServer = FileSystemWatcherServer_1 = tslib_1.__decorate([
    (0, di_1.Injectable)({ multiple: true }),
    tslib_1.__param(0, (0, di_1.Optional)()),
    tslib_1.__metadata("design:paramtypes", [Array])
], FileSystemWatcherServer);
exports.FileSystemWatcherServer = FileSystemWatcherServer;
//# sourceMappingURL=file-service-watcher.js.map