"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getSafeFileservice = exports.FileService = exports.FileSystemNodeOptions = void 0;
const tslib_1 = require("tslib");
const os_1 = tslib_1.__importDefault(require("os"));
const path_1 = tslib_1.__importDefault(require("path"));
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const di_1 = require("@opensumi/di");
const ide_core_node_1 = require("@opensumi/ide-core-node");
const common_1 = require("../common");
const file_system_manage_1 = require("./file-system-manage");
const file_type_1 = require("./shared/file-type");
class FileSystemNodeOptions {
}
exports.FileSystemNodeOptions = FileSystemNodeOptions;
FileSystemNodeOptions.DEFAULT = {
    encoding: 'utf8',
    overwrite: false,
    recursive: true,
    moveToTrash: true,
};
let FileService = class FileService {
    constructor(options) {
        this.options = options;
        this.watcherId = 0;
        this.watcherDisposerMap = new Map();
        this.watcherWithSchemaMap = new Map();
        this.onFileChangedEmitter = new ide_core_node_1.Emitter();
        this.fileSystemManage = new file_system_manage_1.FileSystemManage();
        this.onFilesChanged = this.onFileChangedEmitter.event;
        this.toDisposable = new ide_core_node_1.DisposableCollection();
        this.filesExcludes = [];
        this.filesExcludesMatcherList = [];
        this.workspaceRoots = [];
        this.getEncodingInfo = ide_core_node_1.getEncodingInfo;
        this.initProvider();
    }
    registerProvider(scheme, provider) {
        this.toDisposable.push(this.fileSystemManage.add(scheme, provider));
        this.toDisposable.push(provider.onDidChangeFile((e) => this.fireFilesChange(e)));
        this.toDisposable.push({
            dispose: () => {
                (this.watcherWithSchemaMap.get(scheme) || []).forEach((id) => this.unwatchFileChanges(id));
            },
        });
        return this.toDisposable;
    }
    async watchFileChanges(uri, options) {
        const id = this.watcherId++;
        const _uri = this.getUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const schemaWatchIdList = this.watcherWithSchemaMap.get(_uri.scheme) || [];
        const watcherId = await provider.watch(_uri.codeUri, {
            excludes: (options && options.excludes) || [],
        });
        this.watcherDisposerMap.set(id, {
            dispose: () => provider.unwatch(watcherId),
        });
        schemaWatchIdList.push(id);
        this.watcherWithSchemaMap.set(_uri.scheme, schemaWatchIdList);
        return id;
    }
    async unwatchFileChanges(watcherId) {
        const disposable = this.watcherDisposerMap.get(watcherId);
        if (!disposable || !disposable.dispose) {
            return;
        }
        disposable.dispose();
    }
    setWatchFileExcludes(excludes) {
        const provider = this.getProvider('file');
        provider.setWatchFileExcludes(excludes);
    }
    getWatchFileExcludes() {
        const provider = this.getProvider('file');
        return provider.getWatchFileExcludes();
    }
    setFilesExcludes(excludes, roots) {
        this.filesExcludes = excludes;
        this.filesExcludesMatcherList = [];
        if (roots) {
            this.setWorkspaceRoots(roots);
        }
        this.updateExcludeMatcher();
    }
    getFilesExcludes() {
        return this.filesExcludes;
    }
    setWorkspaceRoots(roots) {
        this.workspaceRoots = roots;
        this.updateExcludeMatcher();
    }
    async getFileStat(uri) {
        const _uri = this.getUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        try {
            const stat = await provider.stat(_uri.codeUri);
            if (!stat) {
                return undefined;
            }
            return this.filterStat(stat);
        }
        catch (err) {
            if (common_1.FileSystemError.FileNotFound.is(err)) {
                return undefined;
            }
        }
    }
    async exists(uri) {
        this.logger.warn('Deprecated Warning: fileService.exists is deprecated, please use fileService.access instead!');
        return this.access(uri);
    }
    // 后端仍然保留 encoding 能力
    async resolveContent(uri, options) {
        const _uri = this.getUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const stat = await provider.stat(_uri.codeUri);
        if (!stat) {
            throw common_1.FileSystemError.FileNotFound(uri, 'File not found.');
        }
        if (stat.isDirectory) {
            throw common_1.FileSystemError.FileIsADirectory(uri, 'Cannot resolve the content.');
        }
        const encoding = await this.doGetEncoding(options);
        const buffer = await provider.readFile(_uri.codeUri);
        return { stat, content: (0, ide_core_node_1.iconvDecode)(this.getNodeBuffer(buffer), encoding) };
    }
    async setContent(file, content, options) {
        const _uri = this.getUri(file.uri);
        const provider = await this.getProvider(_uri.scheme);
        const stat = await provider.stat(_uri.codeUri);
        if (!stat) {
            throw common_1.FileSystemError.FileNotFound(file.uri, 'File not found.');
        }
        if (stat.isDirectory) {
            throw common_1.FileSystemError.FileIsADirectory(file.uri, 'Cannot set the content.');
        }
        if (!(await this.isInSync(file, stat))) {
            throw this.createOutOfSyncError(file, stat);
        }
        const encoding = await this.doGetEncoding(options);
        const buffer = (0, ide_core_node_1.iconvEncode)(content, encoding);
        await provider.writeFile(_uri.codeUri, buffer, { create: false, overwrite: true, encoding });
        const newStat = await provider.stat(_uri.codeUri);
        if (!newStat) {
            throw common_1.FileSystemError.FileNotFound(_uri.codeUri.toString(), 'File not found.');
        }
        return newStat;
    }
    async updateContent(file, contentChanges, options) {
        const _uri = this.getUri(file.uri);
        const provider = await this.getProvider(_uri.scheme);
        const stat = await provider.stat(_uri.codeUri);
        if (!stat) {
            throw common_1.FileSystemError.FileNotFound(file.uri, 'File not found.');
        }
        if (stat.isDirectory) {
            throw common_1.FileSystemError.FileIsADirectory(file.uri, 'Cannot set the content.');
        }
        if (!this.checkInSync(file, stat)) {
            throw this.createOutOfSyncError(file, stat);
        }
        if (contentChanges.length === 0) {
            return stat;
        }
        const encoding = await this.doGetEncoding(options);
        // const content = await fs.readFile(FileUri.fsPath(_uri), { encoding });
        const buffer = await this.getNodeBuffer(await provider.readFile(_uri.codeUri));
        const content = (0, ide_core_node_1.iconvDecode)(buffer, encoding);
        const newContent = this.applyContentChanges(content, contentChanges);
        const newBuffer = (0, ide_core_node_1.iconvEncode)(newContent, encoding);
        await provider.writeFile(_uri.codeUri, newBuffer, { create: false, overwrite: true, encoding });
        const newStat = await provider.stat(_uri.codeUri);
        if (!newStat) {
            throw common_1.FileSystemError.FileNotFound(_uri.codeUri.toString(), 'File not found.');
        }
        return newStat;
    }
    async move(sourceUri, targetUri, options) {
        const _sourceUri = this.getUri(sourceUri);
        const _targetUri = this.getUri(targetUri);
        const provider = await this.getProvider(_sourceUri.scheme);
        const result = await provider.rename(_sourceUri.codeUri, _targetUri.codeUri, {
            overwrite: !!(options && options.overwrite),
        });
        if (result) {
            return result;
        }
        const stat = await provider.stat(_targetUri.codeUri);
        return stat;
    }
    async copy(sourceUri, targetUri, options) {
        const _sourceUri = this.getUri(sourceUri);
        const _targetUri = this.getUri(targetUri);
        const provider = await this.getProvider(_sourceUri.scheme);
        const overwrite = await this.doGetOverwrite(options);
        if (!(0, common_1.containsExtraFileMethod)(provider, 'copy')) {
            throw this.getErrorProvideNotSupport(_sourceUri.scheme, 'copy');
        }
        const result = await provider.copy(_sourceUri.codeUri, _targetUri.codeUri, {
            overwrite: !!overwrite,
        });
        if (result) {
            return result;
        }
        const stat = await provider.stat(_targetUri.codeUri);
        return stat;
    }
    async createFile(uri, options = {}) {
        const _uri = this.getUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const content = await this.doGetContent(options);
        const encoding = await this.doGetEncoding(options);
        const buffer = (0, ide_core_node_1.iconvEncode)(content, encoding);
        let newStat = await provider.writeFile(_uri.codeUri, buffer, {
            create: true,
            overwrite: (options && options.overwrite) || false,
            encoding,
        });
        newStat = newStat || (await provider.stat(_uri.codeUri));
        return newStat;
    }
    async createFolder(uri) {
        const _uri = this.getUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        const result = await provider.createDirectory(_uri.codeUri);
        if (result) {
            return result;
        }
        const stat = await provider.stat(_uri.codeUri);
        return stat;
    }
    /**
     * @param {string} uri
     * @param {FileDeleteOptions} [options] Only support scheme `file`
     * @returns {Promise<void>}
     * @memberof FileService
     */
    async delete(uri, options) {
        const _uri = this.getUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        await provider.stat(_uri.codeUri);
        await provider.delete(_uri.codeUri, {
            recursive: true,
            moveToTrash: await this.doGetMoveToTrash(options),
        });
    }
    async access(uri, mode = common_1.FileAccess.Constants.F_OK) {
        const _uri = this.getUri(uri);
        const provider = await this.getProvider(_uri.scheme);
        if (!(0, common_1.containsExtraFileMethod)(provider, 'access')) {
            throw this.getErrorProvideNotSupport(_uri.scheme, 'access');
        }
        return await provider.access(_uri.codeUri, mode);
    }
    /**
     * Only support scheme `file`
     * @param {string} uri
     * @returns {Promise<string>}
     * @memberof FileService
     */
    async getEncoding(uri) {
        // TODO: 临时修复方案 目前识别率太低，全部返回 UTF8
        return ide_core_node_1.UTF8;
        // const _uri = this.getUri(uri);
        // if (_uri.scheme !== Schemas.file) {
        //   console.warn(`Only support scheme file!, will return UTF8!`);
        //   return UTF8;
        // }
        // const provider = await this.getProvider(_uri.scheme);
        // const stat = await provider.stat(_uri.codeUri);
        // if (!stat) {
        //   throw FileSystemError.FileNotFound(uri);
        // }
        // if (stat.isDirectory) {
        //   throw FileSystemError.FileIsADirectory(uri, 'Cannot get the encoding.');
        // }
        // const encoding = detectEncodingByURI(_uri);
        // return encoding || this.options.encoding || UTF8;
    }
    async getCurrentUserHome() {
        return this.getFileStat(ide_core_node_1.FileUri.create(os_1.default.homedir()).toString());
    }
    /**
     *
     * Only support scheme `file`
     */
    async getFsPath(uri) {
        if (!uri.startsWith('file:/')) {
            return undefined;
        }
        else {
            return ide_core_node_1.FileUri.fsPath(uri);
        }
    }
    async getFileType(uri) {
        return (0, file_type_1.getFileType)(uri);
    }
    getUri(uri) {
        const _uri = new ide_core_node_1.URI(uri);
        if (!_uri.scheme) {
            throw new Error(`Unsupported to get Uri from non-scheme Uri: ${uri}`);
        }
        return _uri;
    }
    /**
     * Current policy: sends * all * Provider onDidChangeFile events to * all * clients and listeners
     */
    fireFilesChange(e) {
        if (e.length < 1) {
            return false;
        }
        this.logger.debug('FileChangeEvent:', e);
        this.onFileChangedEmitter.fire({
            changes: e,
        });
    }
    dispose() {
        this.toDisposable.dispose();
    }
    updateExcludeMatcher() {
        this.filesExcludes.forEach((str) => {
            if (this.workspaceRoots.length > 0) {
                this.workspaceRoots.forEach((root) => {
                    const uri = new ide_core_node_1.URI(root);
                    const pathStrWithExclude = uri.resolve(str).path.toString();
                    this.filesExcludesMatcherList.push((0, ide_core_node_1.parseGlob)(pathStrWithExclude));
                });
            }
            else {
                this.filesExcludesMatcherList.push((0, ide_core_node_1.parseGlob)(str));
            }
        });
    }
    isExclude(uriString) {
        const uri = new ide_core_node_1.URI(uriString);
        return this.filesExcludesMatcherList.some((matcher) => matcher(uri.path.toString()));
    }
    filterStat(stat) {
        if (!stat) {
            return;
        }
        if (this.isExclude(stat.uri)) {
            return;
        }
        if (stat.children) {
            stat.children = this.filterStatChildren(stat.children);
        }
        return stat;
    }
    filterStatChildren(children) {
        const list = [];
        children.forEach((child) => {
            if (this.isExclude(child.uri)) {
                return false;
            }
            const state = this.filterStat(child);
            if (state) {
                list.push(state);
            }
        });
        return list;
    }
    getNodeBuffer(asBuffer) {
        if (Buffer.isBuffer(asBuffer)) {
            return asBuffer;
        }
        if ((0, ide_core_node_1.isArray)(asBuffer)) {
            return Buffer.from(asBuffer);
        }
        if (asBuffer && (0, ide_core_node_1.isArray)(asBuffer.data)) {
            return Buffer.from(asBuffer.data);
        }
        if (!asBuffer || (0, ide_core_node_1.isEmptyObject)(asBuffer)) {
            return Buffer.from([]);
        }
        return asBuffer;
    }
    getErrorProvideNotSupport(scheme, funName) {
        return `Scheme ${scheme} not support this function: ${funName}.`;
    }
    initProvider() {
        this.registerProvider(ide_core_node_1.Schemes.file, this.injector.get(common_1.IDiskFileProvider));
    }
    getProvider(scheme) {
        const provider = this.fileSystemManage.get(scheme);
        if (!provider) {
            throw new Error(`Not find ${scheme} provider.`);
        }
        return provider;
    }
    applyContentChanges(content, contentChanges) {
        let document = vscode_languageserver_types_1.TextDocument.create('', '', 1, content);
        for (const change of contentChanges) {
            let newContent = change.text;
            if (change.range) {
                const start = document.offsetAt(change.range.start);
                const end = document.offsetAt(change.range.end);
                newContent = document.getText().substr(0, start) + change.text + document.getText().substr(end);
            }
            document = vscode_languageserver_types_1.TextDocument.create(document.uri, document.languageId, document.version, newContent);
        }
        return document.getText();
    }
    async isInSync(file, stat) {
        if (this.checkInSync(file, stat)) {
            return true;
        }
        return false;
    }
    checkInSync(file, stat) {
        return stat.lastModification === file.lastModification && stat.size === file.size;
    }
    createOutOfSyncError(file, stat) {
        return common_1.FileSystemError.FileIsOutOfSync(file, stat);
    }
    /**
     * Filters hidden and system partitions.
     */
    filterMountpointPath(path) {
        // OS X: This is your sleep-image. When your Mac goes to sleep it writes the contents of its memory to the hard disk. (https://bit.ly/2R6cztl)
        if (path === '/private/var/vm') {
            return false;
        }
        // Ubuntu: This system partition is simply the boot partition created when the computers mother board runs UEFI rather than BIOS. (https://bit.ly/2N5duHr)
        if (path === '/boot/efi') {
            return false;
        }
        return true;
    }
    async doGetEncoding(option) {
        return option && typeof option.encoding !== 'undefined' ? option.encoding : this.options.encoding;
    }
    async doGetOverwrite(option) {
        return option && typeof option.overwrite !== 'undefined' ? option.overwrite : this.options.overwrite;
    }
    async doGetRecursive(option) {
        return option && typeof option.recursive !== 'undefined' ? option.recursive : this.options.recursive;
    }
    async doGetMoveToTrash(option) {
        return option && typeof option.moveToTrash !== 'undefined' ? option.moveToTrash : this.options.moveToTrash;
    }
    async doGetContent(option) {
        return (option && option.content) || '';
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_node_1.INodeLogger),
    tslib_1.__metadata("design:type", Object)
], FileService.prototype, "logger", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(di_1.INJECTOR_TOKEN),
    tslib_1.__metadata("design:type", di_1.Injector)
], FileService.prototype, "injector", void 0);
FileService = tslib_1.__decorate([
    (0, di_1.Injectable)(),
    tslib_1.__param(0, (0, di_1.Inject)('FileServiceOptions')),
    tslib_1.__metadata("design:paramtypes", [FileSystemNodeOptions])
], FileService);
exports.FileService = FileService;
// 每一个后端连接对应一个 injector
const safeFsInstanceMap = new Map();
function getSafeFileservice(injector) {
    if (safeFsInstanceMap.get(injector)) {
        return safeFsInstanceMap.get(injector);
    }
    const fileService = injector.get(FileService, [FileSystemNodeOptions.DEFAULT]);
    const appConfig = injector.get(ide_core_node_1.AppConfig);
    fileServiceInterceptor(fileService, [
        'exists',
        'resolveContent',
        'setContent',
        'updateContent',
        'createFile',
        'createFolder',
        'delete',
        'access',
        'getFsPath',
        'getFileType',
        'getFileStat',
        'move',
        'copy',
    ], appConfig.blockPatterns || []);
    safeFsInstanceMap.set(injector, fileService);
    return fileService;
}
exports.getSafeFileservice = getSafeFileservice;
// 对于首个参数为uri的方法进行安全拦截
function fileServiceInterceptor(fileService, blackList, blockPatterns) {
    for (const method of blackList) {
        if (typeof fileService[method] === 'function') {
            const originFunc = fileService[method];
            fileService[method] = (...args) => {
                // 第一个参数为uri/{uri}
                if (blockPatterns.length > 0) {
                    const uri = typeof args[0] === 'string' ? args[0] : args[0].uri;
                    if (typeof uri === 'string') {
                        let resolvedURI = uri;
                        if (uri.startsWith('file://')) {
                            /**
                             * match('file:///test/folder/**', 'file:///test/foo/../test/token')
                             *
                             * 防止被绕过
                             */
                            resolvedURI = `file://${path_1.default.resolve(uri.slice(7))}`;
                        }
                        for (const blockPattern of blockPatterns) {
                            if ((0, ide_core_node_1.match)(blockPattern, resolvedURI)) {
                                throw new Error('illegal accessing ' + uri + ' with rule ' + blockPattern);
                            }
                        }
                    }
                }
                return originFunc.apply(fileService, args);
                // copy和move第二个参数也为uri，只禁止来源
            };
        }
    }
}
//# sourceMappingURL=file-service.js.map